import argparse
import os
import json
import sys
import time

import dotenv
dotenv.load_dotenv('.env')

# Import necessary modules from buelon
import buelon as bue


def delete_last_line():
    sys.stdout.write("\033[F")
    sys.stdout.write("\033[K")


def run_hub(args):
    # Set environment variables for hub
    if args.binding:
        os.environ['PIPELINE_HOST'], os.environ['PIPELINE_PORT'] = args.binding.split(':')

    if args.bucket_binding:
        os.environ['BUCKET_CLIENT_HOST'], os.environ['BUCKET_CLIENT_PORT'] = args.bucket_binding.split(':')

    # Set PostgreSQL environment variables if provided
    if args.postgres:
        os.environ['POSTGRES_HOST'], os.environ['POSTGRES_PORT'], os.environ['POSTGRES_USER'], os.environ['POSTGRES_PASSWORD'], os.environ['POSTGRES_DATABASE'] = args.postgres.split(':')

    # Run the hub
    bue.hub.main()


def run_worker(args):
    # Set environment variables for worker
    if args.binding:
        os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = args.binding.split(':')

    if args.bucket_binding:
        os.environ['BUCKET_CLIENT_HOST'], os.environ['BUCKET_CLIENT_PORT'] = args.bucket_binding.split(':')

    # Set PIPE_WORKER_SUBPROCESS_JOBS environment variable if provided
    if args.subprocess_jobs:
        os.environ['PIPE_WORKER_SUBPROCESS_JOBS'] = args.subprocess_jobs

    # Set PIPE_WORKER_SCOPES environment variable if provided
    if args.scopes:
        os.environ['PIPE_WORKER_SCOPES'] = args.scopes

    # Set PostgreSQL environment variables if provided
    if args.postgres:
        os.environ['POSTGRES_HOST'], os.environ['POSTGRES_PORT'], os.environ['POSTGRES_USER'], os.environ['POSTGRES_PASSWORD'], os.environ['POSTGRES_DATABASE'] = args.postgres.split(':')

    # Run the worker
    bue.worker.main()


def run_bucket(args):
    # Set environment variables for bucket
    if args.binding:
        os.environ['BUCKET_SERVER_HOST'], os.environ['BUCKET_SERVER_PORT'] = args.binding.split(':')

    # Run the bucket
    bue.bucket.main()

def run_demo():
    # Run the demo
    bue.examples.demo.main()


def run_example():
    # Run the example
    bue.examples.example.setup()


def upload_pipe_code(file_path, binding, lazy_steps):
    if binding:
        os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = binding.split(':')
    bue.hub.upload_pipe_code_from_file(file_path, lazy_steps)


def display_status(args):
    if args.binding:
        os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = args.binding.split(':')
    client = bue.hub.HubClient()
    refuse_count = 0
    def display_status_worker():
        nonlocal refuse_count
        try:
            data = client.sync_get_step_count()
        except ConnectionRefusedError:
            print('Connection refused' + '.' * (refuse_count % 3))
            refuse_count += 1
            return 1
        except OSError as e:  # socket error
            print(f'{e}')
            return 1
        client.get_step_count()
        # bue.hub.get_step_count(args.binding)
        for row in data:
            name, amount = row['status'], row['amount']
            print(f'{name}: {amount:,}')
        print(f'total: {sum([row["amount"] for row in data]):,}')
        return len(data) + 1
    if not args.subscribe:
        display_status_worker()
    else:
        last_length = 0

        while True:
            try:
                for _ in range(last_length):
                    delete_last_line()
                last_length = display_status_worker()
                time.sleep(3)
            except KeyboardInterrupt:
                print('\n - Have a great day! :)\n')
                break


def fetch_errors(args):
    if args.binding:
        os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = args.binding.split(':')

    try:
        count = int(args.count)
    except ValueError:
        count = 25

    exclude = args.exclude
    if not isinstance(exclude, str):
        exclude = None
    elif ',' in exclude:
        exclude = exclude.split(',')

    print('exclude', exclude)

    client = bue.hub.HubClient()
    return_value = client.fetch_errors(count, exclude)
    count, total, table = return_value['count'], return_value['total'], return_value['table']
    print(f'Fetched {count} errors of {total} total errors.\n')
    for row in table:
        _id, step, msg, _trace = row['id'], row['step'], row['msg'], row['trace']
        print(f'ID: \n  {_id}')
        print(f'Step: \n  {json.dumps(step, indent=4)}')
        print(f'Message: \n  {msg}')
        print(f'Trace: \n  {_trace}\n\n--**--\n')


def cli():
    parser = argparse.ArgumentParser(description='Buelon command-line interface')
    parser.add_argument('-v', '--version', action='version', version='Buelon 1.0.45-alpha8')
    parser.add_argument('-b', '--binding', help='Binding for uploading pipe code (host:port)')
    parser.add_argument('file_path', nargs='?', default='nothing', help='File path for uploading pipe code')

    subparsers = parser.add_subparsers(title='Commands', dest='command', required=False)

    # Check if environment variables are already set
    hub_binding_required = not ('PIPELINE_HOST' in os.environ and 'PIPELINE_PORT' in os.environ)
    hub_bucket_required = not ('BUCKET_CLIENT_HOST' in os.environ and 'BUCKET_CLIENT_PORT' in os.environ)

    worker_binding_required = not ('PIPE_WORKER_HOST' in os.environ and 'PIPE_WORKER_PORT' in os.environ)
    worker_bucket_required = not ('BUCKET_CLIENT_HOST' in os.environ and 'BUCKET_CLIENT_PORT' in os.environ)

    bucket_binding_required = not ('BUCKET_SERVER_HOST' in os.environ and 'BUCKET_SERVER_PORT' in os.environ)

    # Pipe file run
    file_parser = subparsers.add_parser('upload', help='Upload pipe code from a file')
    file_parser.add_argument('-b', '--hub-binding', required=worker_binding_required, help='Binding to hub (host:port)')
    file_parser.add_argument('-f', '--file-path', help='File path for uploading pipe code', dest='file_path')
    file_parser.add_argument('-l', '--lazy', default=False, action=argparse.BooleanOptionalAction, dest='lazy')

    # Hub command
    hub_parser = subparsers.add_parser('hub', help='Run the hub')
    hub_parser.add_argument('-b', '--binding', required=hub_binding_required, help='Main binding for hub (host:port)')
    hub_parser.add_argument('-k', '--bucket-binding', required=hub_bucket_required, help='Bucket binding (host:port)')
    hub_parser.add_argument('-p', '--postgres', help='Postgres connection (host:port:user:password:database)')

    # Worker command
    worker_parser = subparsers.add_parser('worker', help='Run the worker')
    worker_parser.add_argument('-b', '--binding', required=worker_binding_required, help='Main binding for worker (host:port)')
    worker_parser.add_argument('-k', '--bucket-binding', required=worker_bucket_required, help='Bucket binding (host:port)')
    worker_parser.add_argument('-j', '--subprocess-jobs', choices=['true', 'false'], help='Set PIPE_WORKER_SUBPROCESS_JOBS environment variable')
    worker_parser.add_argument('-s', '--scopes', help='Set PIPE_WORKER_SCOPES environment variable')
    worker_parser.add_argument('-p', '--postgres', help='Postgres connection (host:port:user:password:database)')

    # Bucket command
    bucket_parser = subparsers.add_parser('bucket', help='Run the bucket')
    bucket_parser.add_argument('-b', '--binding', required=bucket_binding_required, help='Binding for bucket (host:port)')

    # Status
    status_parser = subparsers.add_parser('status', help='Check the status of the pipeline')
    status_parser.add_argument('-b', '--binding', required=worker_binding_required, help='Main binding for hub (host:port)')
    status_parser.add_argument('-s', '--subscribe', default=False, action=argparse.BooleanOptionalAction, help='Subscribe to status updates')

    # Reset Errors
    reset_parser = subparsers.add_parser('reset', help='Reset errors')
    reset_parser.add_argument('-b', '--binding', required=worker_binding_required, help='Main binding for hub (host:port)')
    reset_parser.add_argument('-w', '--include_working', default='false', choices=['true', 'false'], help='Whether to reset status \'working\'')

    # Delete steps
    delete_parser = subparsers.add_parser('delete', help='Delete steps')
    delete_parser.add_argument('-b', '--binding', required=worker_binding_required, help='Main binding for hub (host:port)')
    # delete_parser.add_argument('-s', '--step_id', help='Step ID to delete')

    # Fetch Errors
    error_fetch_parser = subparsers.add_parser('errors', help='View Error Logs')
    error_fetch_parser.add_argument('-b', '--binding', required=worker_binding_required,  help='Main binding for hub (host:port)')
    error_fetch_parser.add_argument('-c', '--count', default='10', help='Amount of error logs to view (must be int)')
    error_fetch_parser.add_argument('-e', '--exclude', default=None, help='Exclude a specific strings from error message. Commas exclude multiple messages')

    # Fetch Errors
    run_step_parser = subparsers.add_parser('run-step', help='View Error Logs')
    run_step_parser.add_argument('-s', '--step', required=True, help='The step id')

    # Demo command
    demo_parser = subparsers.add_parser('demo', help='Run the demo')
    # demo_parser.set_defaults(func=run_demo)

    # Example command
    example_parser = subparsers.add_parser('example', help='Run the example')
    # example_parser.set_defaults(func=run_example)

    # Parse arguments
    args, remaining_args = parser.parse_known_args()
    # Handle the commands
    if args.command == 'hub':
        run_hub(args)
    elif args.command == 'worker':
        run_worker(args)
    elif args.command == 'bucket':
        run_bucket(args)
    elif args.command == 'demo':
        run_demo()
    elif args.command == 'example':
        run_example()
    elif args.command == 'upload':
        upload_pipe_code(args.file_path, args.hub_binding, args.lazy)
    elif args.command == 'reset':
        if args.binding:
            os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = args.binding.split(':')
        # bue.hub.reset_errors(args.include_working is not None)
        client = bue.hub.HubClient()
        client.reset_errors(args.include_working == 'true')
    elif args.command == 'status':
        display_status(args)
    elif args.command == 'delete':
        if args.binding:
            os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = args.binding.split(':')
        client = bue.hub.HubClient()
        client.sync_delete_steps()
    elif args.command == 'errors':
        fetch_errors(args)
    elif args.command == 'run-step':
        if args.binding:
            os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = args.binding.split(':')
        if not os.environ.get('PIPE_WORKER_HOST') or not os.environ.get('PIPE_WORKER_PORT'):
            raise ValueError('Binding is required when running a step. Set env vars `PIPE_WORKER_HOST` and `PIPE_WORKER_PORT`')

        step_id = args.step

        bue.worker.job(step_id)
    else:
        # Handle the case where a file path is given without a command
        if args.binding and remaining_args:
            file_path = remaining_args[0]
            file_path.close()
            # upload_pipe_code(file_path, args.binding)
        elif remaining_args:
            parser.error('Binding is required when uploading pipe code from file. Use -b or --binding.')
        else:
            parser.print_help()
            sys.exit(1)


if __name__ == '__main__':
    cli()



# import argparse
# import os
# import sys
#
# # Import necessary modules from buelon
# import buelon as bue
#
# def run_hub(args):
#     # Set environment variables for hub
#     os.environ['PIPELINE_HOST'], os.environ['PIPELINE_PORT'] = args.main_binding.split(':')
#     os.environ['BUCKET_CLIENT_HOST'], os.environ['BUCKET_CLIENT_PORT'] = args.bucket_binding.split(':')
#
#     # Set PostgreSQL environment variables if provided
#     if args.postgres:
#         os.environ['POSTGRES_HOST'], os.environ['POSTGRES_PORT'], os.environ['POSTGRES_USER'], os.environ['POSTGRES_PASSWORD'], os.environ['POSTGRES_DATABASE'] = args.postgres.split(':')
#
#     # Run the hub
#     bue.hub.main()
#
# def run_worker(args):
#     # Set environment variables for worker
#     os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = args.main_binding.split(':')
#     os.environ['BUCKET_CLIENT_HOST'], os.environ['BUCKET_CLIENT_PORT'] = args.bucket_binding.split(':')
#
#     # Set PIPE_WORKER_SUBPROCESS_JOBS environment variable if provided
#     if args.subprocess_jobs:
#         os.environ['PIPE_WORKER_SUBPROCESS_JOBS'] = args.subprocess_jobs
#
#     # Set PIPE_WORKER_SCOPES environment variable if provided
#     if args.scopes:
#         os.environ['PIPE_WORKER_SCOPES'] = args.scopes
#
#     # Set PostgreSQL environment variables if provided
#     if args.postgres:
#         os.environ['POSTGRES_HOST'], os.environ['POSTGRES_PORT'], os.environ['POSTGRES_USER'], os.environ['POSTGRES_PASSWORD'], os.environ['POSTGRES_DATABASE'] = args.postgres.split(':')
#
#     # Run the worker
#     bue.worker.main()
#
# def run_bucket(args):
#     # Set environment variables for bucket
#     os.environ['BUCKET_SERVER_HOST'], os.environ['BUCKET_SERVER_PORT'] = args.binding.split(':')
#
#     # Run the bucket
#     bue.bucket.main()
#
# def run_demo():
#     # Run the demo
#     bue.examples.demo.main()
#
# def run_example():
#     # Run the example
#     bue.examples.example.main()
#
# def upload_pipe_code(file_path, binding):
#     os.environ['PIPE_WORKER_HOST'], os.environ['PIPE_WORKER_PORT'] = binding.split(':')
#     bue.hub.upload_pipe_code_from_file(file_path)
#
# def cli():
#     parser = argparse.ArgumentParser(description='Buelon command-line interface')
#     parser.add_argument('-v', '--version', action='version', version='Buelon 1.0.0')
#     subparsers = parser.add_subparsers(title='Commands', dest='command')
#
#     # Check if environment variables are already set
#     hub_binding_required = not ('PIPELINE_HOST' in os.environ and 'PIPELINE_PORT' in os.environ)
#     hub_bucket_required = not ('BUCKET_CLIENT_HOST' in os.environ and 'BUCKET_CLIENT_PORT' in os.environ)
#
#     worker_binding_required = not ('PIPE_WORKER_HOST' in os.environ and 'PIPE_WORKER_PORT' in os.environ)
#     worker_bucket_required = not ('BUCKET_CLIENT_HOST' in os.environ and 'BUCKET_CLIENT_PORT' in os.environ)
#
#     bucket_binding_required = not ('BUCKET_SERVER_HOST' in os.environ and 'BUCKET_SERVER_PORT' in os.environ)
#
#     # Hub command
#     hub_parser = subparsers.add_parser('hub', help='Run the hub')
#     hub_parser.add_argument('-b', '--main-binding', required=hub_binding_required, help='Main binding for hub (host:port)')
#     hub_parser.add_argument('-k', '--bucket-binding', required=hub_bucket_required, help='Bucket binding (host:port)')
#     hub_parser.add_argument('-p', '--postgres', help='Postgres connection (host:port:user:password:database)')
#
#     # Worker command
#     worker_parser = subparsers.add_parser('worker', help='Run the worker')
#     worker_parser.add_argument('-b', '--main-binding', required=worker_binding_required, help='Main binding for worker (host:port)')
#     worker_parser.add_argument('-k', '--bucket-binding', required=worker_bucket_required, help='Bucket binding (host:port)')
#     worker_parser.add_argument('-s', '--subprocess-jobs', choices=['true', 'false'], help='Set PIPE_WORKER_SUBPROCESS_JOBS environment variable')
#     worker_parser.add_argument('--scopes', help='Set PIPE_WORKER_SCOPES environment variable')
#     worker_parser.add_argument('-p', '--postgres', help='Postgres connection (host:port:user:password:database)')
#
#     # Bucket command
#     bucket_parser = subparsers.add_parser('bucket', help='Run the bucket')
#     bucket_parser.add_argument('-b', '--binding', required=bucket_binding_required, help='Binding for bucket (host:port)')
#
#     # Demo command
#     demo_parser = subparsers.add_parser('demo', help='Run the demo')
#     demo_parser.set_defaults(func=run_demo)
#
#     # Example command
#     example_parser = subparsers.add_parser('example', help='Run the example')
#     example_parser.set_defaults(func=run_example)
#
#     # Parse arguments
#     args, remaining_args = parser.parse_known_args()
#
#     # Handle the commands
#     if args.command == 'hub':
#         run_hub(args)
#     elif args.command == 'worker':
#         run_worker(args)
#     elif args.command == 'bucket':
#         run_bucket(args)
#     elif args.command == 'demo':
#         run_demo()
#     elif args.command == 'example':
#         run_example()
#     else:
#         # Handle the case where a file path is given without a command
#         if remaining_args:
#             file_path = remaining_args[-1]
#             if '--binding' in remaining_args or '-b' in remaining_args:
#                 index = remaining_args.index('--binding') + 1 if '--binding' in remaining_args else remaining_args.index('-b') + 1
#                 binding = remaining_args[index]
#                 upload_pipe_code(file_path, binding)
#             else:
#                 parser.error('Binding is required when uploading pipe code from file.')
#         else:
#             parser.print_help()
#             sys.exit(1)
#
# if __name__ == '__main__':
#     cli()
#



import os
import asyncio
import traceback
import sys
import enum
import contextlib

import asyncio_pool

import buelon.core.step
import buelon.hub
import buelon.bucket
import buelon.helpers.json_parser

import time

try:
    import dotenv
    dotenv.load_dotenv('.env')
except ModuleNotFoundError:
    pass

DEFAULT_SCOPES = 'production-heavy,production-medium,production-small,testing-heavy,testing-medium,testing-small,default'

WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
PIPE_WORKER_SUBPROCESS_JOBS = os.environ.get('PIPE_WORKER_SUBPROCESS_JOBS', 'true')
try:
    N_WORKER_PROCESSES: int = int(os.environ['N_WORKER_PROCESSES'])
except (KeyError, ValueError):
    N_WORKER_PROCESSES = 15

bucket_client = buelon.bucket.Client()
hub_client: buelon.hub.HubClient = buelon.hub.HubClient(WORKER_HOST, WORKER_PORT)

JOB_CMD = f'{sys.executable} -c "import buelon.worker;buelon.worker.job()"'

TEMP_FILE_LIFETIME = 60 * 60 * 3


class HandleStatus(enum.Enum):
    success = 'success'
    pending = 'pending'
    almost = 'almost'
    none = 'none'


@contextlib.contextmanager
def new_client_if_subprocess():
    global hub_client
    if PIPE_WORKER_SUBPROCESS_JOBS == 'true':
        with buelon.hub.HubClient(WORKER_HOST, WORKER_PORT) as client:
            yield client
    else:
        yield hub_client


def job(step_id: str | None = None) -> None:
    if step_id:
        _step = buelon.hub.get_step(step_id)
    else:
        _step = buelon.hub.get_step(os.environ['STEP_ID'])

    if _step is None:
        with new_client_if_subprocess() as client:
            client.reset(step_id if step_id else os.environ['STEP_ID'])
            return

    print('handling', _step.name)
    try:
        args = [buelon.hub.get_data(_id) for _id in _step.parents]
        r: buelon.core.step.Result = _step.run(*args)
        buelon.hub.set_data(_step.id, r.data)

        with new_client_if_subprocess() as client:
            client: buelon.hub.HubClient
            if r.status == buelon.core.step.StepStatus.success:
                client.done(_step.id)
            elif r.status == buelon.core.step.StepStatus.pending:
                client.pending(_step.id)
            elif r.status == buelon.core.step.StepStatus.reset:
                client.reset(_step.id)
            elif r.status == buelon.core.step.StepStatus.cancel:
                client.cancel(_step.id)
            else:
                raise Exception('Invalid step status')
    except Exception as e:
        print(' - Error - ')
        print(str(e))
        traceback.print_exc()
        with new_client_if_subprocess() as client:
            client.error(
                _step.id,
                str(e),
                f'{traceback.format_exc()}'
            )


async def run(step_id: str | None = None) -> None:
    if PIPE_WORKER_SUBPROCESS_JOBS != 'true':
        return job(step_id)
    env = {**os.environ, 'STEP_ID': step_id}
    p = await asyncio.create_subprocess_shell(JOB_CMD, env=env)
    await p.wait()


async def work():
    _scopes: str = os.environ.get('PIPE_WORKER_SCOPES', DEFAULT_SCOPES)
    scopes: list[str] = _scopes.split(',')
    print('scopes', scopes)

    last_loop_had_steps = True

    async with asyncio_pool.AioPool(size=N_WORKER_PROCESSES) as pool:
        while True:
            steps = hub_client.get_steps(scopes)

            if not steps:
                if last_loop_had_steps:
                    last_loop_had_steps = False
                    print('waiting..')
                await asyncio.sleep(1.)
                continue

            last_loop_had_steps = True

            for s in steps:
                await pool.spawn(run(s))


def is_hanging_script(path: str):
    """
    Checks if a temporary script created by the worker that was not properly cleaned up

    Args:
        path: file path

    Returns: True if the file is a hanging script
    """
    # example: temp_ace431278698111efab2de73d545b8b66.py
    file_name = os.path.basename(path)
    return (file_name.startswith('temp_')
            and file_name.endswith('.py')
            # and len(file_name) == 41
            and (time.time() - os.path.getmtime(path)) > TEMP_FILE_LIFETIME)


async def cleaner():
    """
    Cleans up hanging scripts created by the worker that were not properly cleaned up
    """
    for root, dirs, files in os.walk('.'):
        for file in files:
            if is_hanging_script(os.path.join(root, file)):
                os.remove(os.path.join(root, file))
            await asyncio.sleep(0.1)
        break


def main():
    """
    Main function to run the worker
    """
    asyncio.run(_main())


async def _main():
    """
    Main coroutine to run the worker
    """
    global hub_client
    asyncio.create_task(cleaner())
    with hub_client:
        await work()


if __name__ == '__main__':
    main()



"""
This module provides functionality for evaluating pipes and handling loops in a custom pipeline language.
"""

import inspect
from typing import Callable

from . import step_definition
from . import pipe
from buelon.helpers import pipe_util
from . import pipe_interpreter


def pipe_eval(index: int, txt: str, variables: dict):
    """
    Evaluate a pipe expression and execute its steps.

    Args:
        index (int): The current line index in the pipeline code.
        txt (str): The pipe expression to evaluate.
        variables (dict): A dictionary containing pipeline variables and functions.

    Returns:
        The result of executing the pipe's steps.

    Raises:
        SyntaxError: If the function call is invalid or unknown.
        TypeError: If the called object is not a pipe.Pipe instance.
    """
    if txt.count('(') == 1 and txt.endswith(')'):
        name, args = txt[:-1].split('(')
        name = pipe_util.trim(name)
        if name not in variables:
            raise SyntaxError(f'Line {index+1}: call to unknown function \'{name}\'')
        _pipe: pipe.Pipe = variables[name]
        if not isinstance(_pipe, pipe.Pipe):
            raise TypeError(f'Line {index+1}: \'{name}\' is not a pipe.Pipe. It\'s \'{type(_pipe)}\'')

        steps = _pipe.create_steps(index, variables, args, pipe_util.json_copy(_pipe.kwargs))

        data = []

        for step_id in steps:
            s = variables['__steps__'][step_id]
            new_data = s.run(data).data
            del data
            del s
            data = new_data

        for step_id in steps:
            del variables['__steps__'][step_id]
            ss = variables['__starters__']
            ss.remove(step_id)
            variables['__starters__'] = ss

        return data


def check_loop(
        index: int,
        line: str,
        lines: list[str],
        variables: dict,
        read_line: Callable[[int], None],
        set_index: Callable[[int], None]
) -> bool:
    """
    Check and process a loop construct in the pipeline code.

    Args:
        index (int): The current line index in the pipeline code.
        line (str): The current line of code being processed.
        lines (list[str]): All lines of the pipeline code.
        variables (dict): A dictionary containing pipeline variables and functions.
        read_line (Callable[[int], None]): A function to process a specific line of code.
        set_index (Callable[[int], None]): A function to set the current line index.

    Returns:
        bool: True if a loop was found and processed, False otherwise.

    Raises:
        TypeError: If the loop variable is not a list.
    """
    if (line.strip().startswith('for ')
            and line.count(' in ') == 1
            and line.strip().endswith(':')):
        before, after = line.split(' in ')
        var_name = pipe_util.trim(before.strip()[3:])
        loop_var = pipe_util.trim(after.strip()[:-1])

        loop_val = pipe_eval(index, loop_var, variables)
        if '(' in loop_var:
            loop_var = loop_var[:loop_var.find('(')]

        if not isinstance(loop_val, list) and not inspect.isgenerator(loop_val):
            raise TypeError(f'Line {index + 1}: Loop variable must be a list or generator. Is {type(loop_val).__name__}')

        def get_indentation(text: str):
            """
            Calculate the indentation level of a line of code.

            Args:
                text (str): A line of code.

            Returns:
                int: The indentation level.
            """
            indentation = 0
            while text.startswith(pipe_interpreter.INDENT):
                indentation += 1
                text = text[len(pipe_interpreter.INDENT):]
            return indentation

        indentation = get_indentation(line)
        next_line = index + 1
        future_lines = []
        for j in range(index + 1, len(lines)):
            if get_indentation(lines[j]) <= indentation:
                break
            future_lines.append(j)
            next_line = j + 1

        for _i, value in enumerate(loop_val):
            _code = f'def main(*args):\n    return {value}'
            sd = step_definition.StepDefinition(f'def_{var_name}_{_i}', 'PYTHON', 'main', '', _code, True)
            sd.scope = variables['__scope__']

            # inherent priority, scope and velocity from parent
            loop_pipe = variables.get(loop_var)
            if isinstance(loop_pipe, pipe.Pipe):
                parent_sd_name = loop_pipe.imports.split(',')[-1]
                parent_sd = variables.get(parent_sd_name)
                if isinstance(parent_sd, step_definition.StepDefinition):
                    sd.priority = parent_sd.priority
                    sd.scope = parent_sd.scope
                    sd.velocity = parent_sd.velocity

            variables[f'def_{var_name}_{_i}'] = sd
            p = pipe.Pipe(var_name, f'def_{var_name}_{_i}')
            variables[var_name] = p.create_steps(index, variables, '')
            for future_line in future_lines:
                read_line(future_line)

            del value

        set_index(next_line - 1)
        return True
    return False

